#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/stat.h>
#include <CL/cl.h>

const unsigned int NUMBERS = 1000;

int main(int argc, char** argv)
{
	cl_int err;
	float input[NUMBERS];
	float output[NUMBERS];
	unsigned int count = NUMBERS;
	
	for (int i = 0; i < count; i++)
	{
		input[i] = i;
	}
 
	// Zuerst muss die Plattform gefunden werden.
	cl_platform_id platform_id;
	unsigned int platforms_count;
	err = clGetPlatformIDs(1, &platform_id, &platforms_count);
	if (err != CL_SUCCESS)
	{
		printf("Fehler: Plattform-IDs konnten nicht gefunden werden!\n");
		return EXIT_FAILURE;
	}

	// Wir suchen ein Compute-Device:
	// CL_DEVICE_TYPE_ALL, CL_DEVICE_TYPE_GPU oder CL_DEVICE_TYPE_CPU
	cl_device_id device_id;
	err = clGetDeviceIDs(platform_id, CL_DEVICE_TYPE_GPU, 1, &device_id, NULL);
	if (err != CL_SUCCESS)
	{
		printf("Fehler: Device-IDs konnten nicht gefunden werden!\n");
		return EXIT_FAILURE;
	}

	// Compute-Context erzeugen.
	cl_context context = clCreateContext(0, 1, &device_id, NULL, NULL, &err);
	if (!context)
	{
		printf("Fehler: Compute-Context konnte nicht erzeugt werden!\n");
		return EXIT_FAILURE;
	}	
	
	// Kommandowarteschlange erzeugen.
	cl_command_queue queue = clCreateCommandQueue(context, device_id, 0, &err);
	if (!queue)
	{
		printf("Fehler: Kommandowarteschlange konnte nicht erzeugt werden!\n");
		return EXIT_FAILURE;
	}

	// Kernel-Quelltext aus Datei lesen und in Buffer kopieren.
	char* kernel_file_name = "wurzel.cl";
	// Datei öffnen
	FILE* kernel_file = fopen(kernel_file_name, "r");
	if (kernel_file == NULL)
	{
		printf("Fehler: Kernelquelltextdatei konnte nicht geöffnet werden!\n");
		return EXIT_FAILURE;
	}
	
	// Dateigröße ermitteln
	struct stat file_stat;
	err = stat(kernel_file_name, &file_stat);
	if (err != 0)
	{
		printf("Fehler: Kernelquelltextdatei konnte nicht gefunden werden!\n");
		return EXIT_FAILURE;
	}
	unsigned int kernel_source_size = file_stat.st_size;
	
	// Speicherbereich anfordern
	char* kernel_source = malloc((size_t) kernel_source_size);
	if (kernel_source == NULL)
	{
		printf("Fehler: Nicht ausreichend Speicher (%u B) für Kernelquelltext!\n", kernel_source_size);
		return EXIT_FAILURE;
	}
	
	// Quelltext kopieren.
	int n_read = fread(kernel_source, 1, kernel_source_size, kernel_file);
	if (n_read != kernel_source_size)
	{
		printf("Fehler: Kernelquelltext konnte nicht vollständig gelesen werden!\n");
		return EXIT_FAILURE;
	}
	fclose(kernel_file);
		
	// Programm aus Kernelquelltext erzeugen
	cl_program program = clCreateProgramWithSource(context, 1, (const char **) &kernel_source, NULL, &err);
	if (!program)
	{
		printf("Fehler: Programm konnte nicht erzeugt werden!\n");
		return EXIT_FAILURE;
	}
	
	// Kernelquelltext übersetzen.
	err = clBuildProgram(program, 0, NULL, NULL, NULL, NULL);
	if (err != CL_SUCCESS)
	{
		size_t len;
		char buffer[2048];

		printf("Fehler beim Übersetzen des Kernelquelltextes!\n");
		clGetProgramBuildInfo(program, device_id, CL_PROGRAM_BUILD_LOG, sizeof(buffer), buffer, &len);
		printf("%s\n", buffer);
		return EXIT_FAILURE;
	}

	// Aus dem übersetzten Quelltext wird jetzt ein „Kernel“ erzeugt.
	cl_kernel kernel = clCreateKernel(program, "square_root", &err);
	if (err != CL_SUCCESS)
	{
		printf("Fehler: Kernel konnte nicht erzeugt werden!\n");
		return EXIT_FAILURE;
	}

	// Allokieren von count float-Speichereinträgen im Device-Speicher, einmal zum Lesen (Eingabe), einmal zum Schreiben (Ausgabe)
	cl_mem input_d = clCreateBuffer(context, CL_MEM_READ_ONLY, sizeof(float) * count, NULL, NULL);
	cl_mem output_d = clCreateBuffer(context, CL_MEM_WRITE_ONLY, sizeof(float) * count, NULL, NULL);
	if (input_d == NULL || output_d == NULL)
	{
		printf("Fehler: Device-Speicher konnte nicht allokiert werden!\n");
		return EXIT_FAILURE;
	}

	// Kopieren des Eingabearrays in den Device-Speicher (Einreihen in die Warteschlange)
	clEnqueueWriteBuffer(queue, input_d, CL_TRUE, 0, sizeof(float) * count, input, 0, NULL, NULL);

	// Argumente für den Aufruf der Kernel-Funktion setzen
	err = clSetKernelArg(kernel, 0, sizeof(cl_mem), &input_d);
	err += clSetKernelArg(kernel, 1, sizeof(cl_mem), &output_d);
	err += clSetKernelArg(kernel, 2, sizeof(unsigned int), &count);
	if (err != CL_SUCCESS)
	{
		printf("Fehler beim Setzen der Kernel-Argumente!\n");
		return EXIT_FAILURE;
	}

	// Einreihen des Aufrufes der Kernel-Funktion in die Warteschlange; pro Block ein Thread; count Blöcke
	size_t global = count;
	size_t local = 1;
	err = clEnqueueNDRangeKernel(queue, kernel, 1, NULL, &global, &local, 0, NULL, NULL);
	if (err != CL_SUCCESS)
	{
		printf("Fehler beim Setzen der Kernel-Argumente!\n");
		return EXIT_FAILURE;
	}

	// Kopieren des Ergebnisearrays aus dem Device-Speicher (Einreihen in die Warteschlange)
	err = clEnqueueReadBuffer(queue, output_d, CL_TRUE, 0, sizeof(float) * count, output, 0, NULL, NULL);
	if (err != CL_SUCCESS)
	{
		printf("Fehler beim Ausführen der Kernel-Funktion!\n");
		return EXIT_FAILURE;
	}
	
	// Warten auf Beendigung der Kommandos in der Warteschlange
	err = clFinish(queue);
	if (err != CL_SUCCESS)
	{
		printf("Fehler beim Warten auf Beendigung!\n");
		return EXIT_FAILURE;
	}

	/* Überprüfung der Werte in output */
	unsigned int correct = 0;
	unsigned int nearly_correct = 0;
	for (int i = 0; i < count; i++)
	{
		if (output[i] == sqrtf(input[i]))
		{
			correct++;
		}
		// Durch unterschiedliche Implementierungen kann es zu minimalen Abweichungen kommen.
		else if (abs(output[i] - sqrtf(input[i])) < 1e-9)
		{
			nearly_correct++;
		}
	}
	printf("%d von %d Wurzeln korrekt berechnet.\n", correct, count);
	printf("%d von %d Wurzeln mit einem Fehler kleiner als 1E-9 korrekt berechnet.\n", correct + nearly_correct, count);
	
	// Freigabe des Grafikspeichers
	clReleaseMemObject(input_d); 
	clReleaseMemObject(output_d); 
	
	// Speicherfreigabe
	clReleaseProgram(program);
	clReleaseKernel(kernel);
	clReleaseCommandQueue(queue);
	clReleaseContext(context);
}
